@@ -11,10 +11,7 @@ import os
 import urllib
 import warnings
 from typing import Union, List
-
-import torch
-from PIL import Image
-from torchvision.transforms import Compose, Resize, CenterCrop, ToTensor, Normalize
+from torchvision.transforms import Compose, Normalize
 from tqdm import tqdm
 
 _MODELS = {
@@ -41,9 +38,14 @@ def _download(url: str, root: str = os.path.expanduser("~/.cache/clip")):
             warnings.warn(f"{download_target} exists, but the SHA256 checksum does not match; re-downloading the file")
 
     with urllib.request.urlopen(url) as source, open(download_target, "wb") as output:
-        with tqdm(total=int(source.info().get("Content-Length")), ncols=80, unit='iB', unit_scale=True) as loop:
+        with tqdm(
+            total=int(source.info().get("Content-Length")),
+            unit='iB',
+            unit_scale=True,
+            desc=f"Downloading {filename}",
+        ) as loop:
             while True:
-                buffer = source.read(8192)
+                buffer = source.read(524288)
                 if not buffer:
                     break
 
@@ -613,9 +615,7 @@ def build_model(state_dict: dict):
     model.load_state_dict(state_dict)
     return model.eval()
 
-import gzip
 import html
-import os
 from functools import lru_cache
 
 import ftfy
@@ -745,6 +745,4 @@ class SimpleTokenizer(object):
         text = ''.join([self.decoder[token] for token in tokens])
         text = bytearray([self.byte_decoder[c] for c in text]).decode('utf-8', errors="replace").replace('</w>', ' ')
         return text
-        import gzip
-
 _tokenizer = SimpleTokenizer()
